# mesonetdownloader.py
#
# Abhiram Pamula (apamula@okstate.edu)
# Ben Rubinstein (brubinst@hawk.iit.edu)
#
# Some code based on nhdplusextractor by Mitchell Sawtelle
#   (https://github.com/Msawtelle/PyNHDPlus)
#
# last updated: 03/03/2023
#
# contains the MesonetDownloader class, which can be used to retrieve and
# manipulate source data from the Mesonet dataset online
import urllib.request, os, pytz
import pandas as pd
import numpy as np
from datetime import timedelta, datetime
from shapefile import Reader
from pyproj import CRS, Transformer
from pandas.errors import EmptyDataError

class MesonetDownloader:
    '''
    The MesonetDownloader class contains methods to generate time series from
        the Oklahoma Mesonet climate dataset.
    '''

    def __init__(self, destination=None, verbose=False):
        '''
        init method for MesonetDownloader class

        arguments:
            destination (str): the path to where the user would like the
                Mesonet data to be stored if no argument is given the location
                of the file being run is used
            verbose (bool): if true write detailed debugging to stdout
        '''
        self.verbose = verbose
        # this is used for handling changing the rain column to differential
        self.errorcodes=[-994,-995,-996,-997,-998,-999]
        # if no destination given in init method set destination to current
        #   directory
        if destination is None:
            self.destination = os.path.dirname(__file__)
        else:
            self.destination = os.path.normpath(destination)

        try:
            if not os.path.isdir(self.destination): os.mkdir(self.destination)
        except PermissionError:
            errstr = ('The chosen destination folder does not exist and you do'
                      ' not have permission to create it.\n{}\nPlease set the'
                      ' destination argument to a different folder.')
            raise PermissionError(errstr.format(self.destination))

        # set up directory structure
        # mts_files holds the downloaded raw data from Mesonet
        # time_series holds saved dataframes as csvs
        self.mts_dir = '{}/{}'.format(self.destination,'mts_files')
        try:
            if not os.path.isdir(self.mts_dir): os.mkdir(self.mts_dir)
        except PermissionError:
            errstr = ('You do not have permission to create the required'
                      ' folders inside the destination directory.\n{}\nPlease'
                      ' change the permissions or choose a different'
                      ' destination folder.')
            raise PermissionError(errstr.format(self.mts_dir))

        # number of times to try in case of failed data retrieval
        self.TTL = 2

        # mesonet base url
        self.base_url = 'http://www.mesonet.org/index.php/'

        # attempt to read metadata from file
        self.metadata = pd.DataFrame()
        try:
            meta_file = 'geoinfo.csv'
            self.metadata = pd.read_csv('{}/{}'.format(self.destination,meta_file))
        except FileNotFoundError:
            if self.verbose:
                print('Metadata file not found locally. Attempting to download'
                      ' from Mesonet.')
        except Exception as e:
            if self.verbose:
                print('Failed to read metadata file locally due to:')
                print(e)
                print('Attempting to download from Mesonet.')
        finally:
            # if we failed to read the file locally, attempt to download
            if self.metadata.empty:
                meta_path = 'api/siteinfo/from_all_active_with_geo_fields/format/csv/'
                meta_url = '{}{}'.format(self.base_url,meta_path)
                attempt = 0
                while attempt < self.TTL:
                    try:
                        urllib.request.urlretrieve(meta_url,'{}/geoinfo.csv'.format(self.destination))
                        self.metadata = pd.read_csv('{}/{}'.format(self.destination,meta_file))
                        break
                    except Exception as e:
                        if self.verbose:
                            print('Attempt number {} to download metadata failed: {}'.format(attempt+1,meta_url))
                            print(e)
                    attempt += 1

        # raise an exception if the metadata file couldn't be read
        if self.metadata.empty:
            raise FileNotFoundError('Metadata file could not be found locally'
                                    ' or downloaded from Mesonet. Try setting'
                                    ' the verbose argument to True for more'
                                    ' details.')

        # convert the commission and decommision columns to datetime objects
        self.metadata['datc'] = pd.to_datetime(self.metadata['datc'],format='%Y%m%d')
        self.metadata['datd'] = pd.to_datetime(self.metadata['datd'],format='%Y%m%d')


    def get_station_ids(self):
        '''
        Returns a list of all the valid station IDs (STIDs) from the metadata file

        returns:
            str: a list of all the station IDs separated by spaces
        '''
        return ' '.join(self.metadata['stid'].values[1:])


    def get_station_coord(self,site_id):
        '''
        Returns the coordinates for a given station ID

        arguments:
            site_id (str): case-insensitive site ID as listed in the metadata
                file

        returns:
            tuple: lat/lon coordinates for the station
        '''
        # check if site id is valid
        if len(self.metadata[self.metadata['stid']==site_id.upper()]) == 0:
            if self.verbose:
                print('Error: invalid site ID. Valid site IDs are:')
                print(' '.join(self.metadata['stid'].values[1:]))
            return None
        lat = self.metadata.loc[self.metadata['stid'] == site_id.upper(),'nlat'].item()
        lon = self.metadata.loc[self.metadata['stid'] == site_id.upper(),'elon'].item()
        return (lat,lon)


    def get_location(self,site_id):
        '''
        Returns the coordinates and elevation for a given station ID

        arguments:
            site_id (str): case-insensitive site ID as listed in the metadata
                file

        returns:
            tuple: latitude, longitude, elevation in meters for the station
        '''
        elev = self.metadata.loc[self.metadata['stid'] == site_id.upper(),'elev'].item()
        return (*self.get_station_coord(site_id), elev)


    def download_station_data(self,site_id,start_date,end_date):
        '''
        Method to download data for a single station over a specified time period

        arguments:
            site_id (str): the ID for the station, IDs are in the metadata file
                case-insensitive
            start_date (datetime.datetime): first date to get data
            end_date (datetime.datetime): last date (inclusive) to get data

        returns:
            DataFrame: a pandas DataFrame containing the downloaded data sorted
                by datetime
        '''
        # note: site ids are in all caps in the metadata file but all lowercase
        #   in the URLs, this ensures that our code is case-insensitive
        site_id = site_id.lower()
        # check if site id is valid
        if len(self.metadata[self.metadata['stid']==site_id.upper()]) == 0:
            if self.verbose:
                print('Error: invalid site ID. Valid site IDs are:')
                print(' '.join(self.metadata['stid'].values[1:]))
            return None
        # check if station was active during the requested time period
        datc = self.metadata.loc[self.metadata['stid'] == site_id.upper(),'datc'].item()
        datd = self.metadata.loc[self.metadata['stid'] == site_id.upper(),'datd'].item()
        if start_date < datc:
            if self.verbose:
                print('Warning: the {} station was commissioned on {} which'
                      ' was after the requested start date of {}. It will be'
                      ' skipped.'.format(site_id.upper(),datc,start_date))
            return None
        if end_date > datd:
            if self.verbose:
                print('Warning: the {} station was decommissioned on {} which'
                      ' was before the requested end date of {}. It will be'
                      ' skipped.'.format(site_id.upper(),datd,end_date))
            return None
        dates_list = []
        # create list of dates to download
        delta = timedelta(days=1)
        while start_date <= end_date:
            dates_list.append(start_date.strftime('%Y%m%d'))
            start_date += delta
        df_list = []
        for i in dates_list:
            cur_df = pd.DataFrame()
            # attempt to read file locally
            try:
                cur_df = pd.read_csv('{}/{}{}.mts'.format(self.mts_dir,i,site_id),header=2,sep=r'\s+')
            except FileNotFoundError:
                if self.verbose:
                    print('File not found locally, attempting to download.')
                    print('{}/{}{}.mts'.format(self.mts_dir,i,site_id))
                attempt = 0
                while attempt < self.TTL:
                    try:
                        url = '{}dataMdfMts/dataController/getFile/{}{}/mts/DOWNLOAD/'.format(self.base_url,i,site_id)
                        urllib.request.urlretrieve(url,'{}/{}{}.mts'.format(self.mts_dir,i,site_id))
                        cur_df = pd.read_csv('{}/{}{}.mts'.format(self.mts_dir,i,site_id),header=2,sep=r'\s+')
                        break
                    except:
                        if self.verbose:
                            print('Failed to download data from {} on attempt {}.'.format(url,attempt+1))
                    attempt += 1
            except EmptyDataError:
                if self.verbose:
                    print('{}/{}{}.mts is an empty file.'.format(self.mts_dir,i,site_id))
                    print('This means that there is no data available for the'
                          ' {} on {}.'.format(site_id,i))
            if not cur_df.empty:
                date =  datetime.strptime(i, '%Y%m%d')
                cur_df['DATETIME'] = cur_df.apply(lambda x: date+timedelta(minutes = int(x['TIME'])), axis = 1)
                # change rain to differential instead of cumulative
                # make a copy of the rain column so we can track the errors
                cur_df['RAIN_BAK'] = cur_df['RAIN']
                # replace the errors with nan
                for code in self.errorcodes:
                    cur_df['RAIN'].replace(str(code),np.nan,inplace=True)
                    cur_df['RAIN'].replace(code,np.nan,inplace=True)
                first_rain = cur_df.loc[0,'RAIN']
                second_rain = cur_df.loc[1,'RAIN']
                # there are cases where the last reading from the previous day
                # rolls over so we check for that and zero out the number
                if first_rain > second_rain:
                    first_rain = 0
                    cur_df.loc[0,'RAIN'] = first_rain
                cur_df['RAIN']=cur_df['RAIN'].diff()
                cur_df.loc[0,'RAIN'] = first_rain
                # copy error codes back into the rain column
                # TODO: fix the SettingWithCopyError warning
                pd.options.mode.chained_assignment = None
                for code in self.errorcodes:
                    cur_df['RAIN'].loc[cur_df['RAIN_BAK']==code] = code
                pd.options.mode.chained_assignment = 'warn'
                # drop some columns to save memory
                cur_df.drop(columns=['STNM','TIME','RAIN_BAK'],inplace=True)
                df_list.append(cur_df)
            elif self.verbose:
                print('No data available for {} on {}.'.format(site_id,i))
        if len(df_list) == 0:
            return None
        final_df = pd.concat(df_list)
        final_df = final_df.set_index('DATETIME')
        return final_df


    def download_bounding_box(self,bbox,start_date,end_date,padding=1):
        '''
        Downloads data for all the stations found within a bounding box.

        arguments:
            bbox (list): the bounding box to be searched; it expects the same format
                as used in pyshp, [low longitude, low latitude, highlon, highlat]
            start_date (datetime.datetime): first date to get data
            end_date (datetime.datetime): last date (inclusive) to get data
            padding (int): an optional parameter that increases or decreases
                the size of the bounding box, must be positive

        returns:
            dict: a dictionary with station IDs as keys and DataFrames holding
                the downloaded data as values
        '''
        # make bbox coordinates easier to work with
        lowlon = bbox[0]
        lowlat = bbox[1]
        highlon = bbox[2]
        highlat = bbox[3]
        if padding != 1 and padding > 0:
            lonpad = padding*(highlon-lowlon)/2
            latpad = padding*(highlat-lowlat)/2
            lowlon = lowlon - lonpad
            lowlat = lowlat - latpad
            highlon = highlon + lonpad
            highlat = highlat + latpad
        elif self.verbose and padding <= 0:
            print('Warning: The padding parameter must be positive. It will be'
                  ' ignored.')
        # find station IDs within the box
        station_ids = []
        for index, row in self.metadata.iterrows():
            lat = row.nlat
            lon = row.elon
            if lat > lowlat and lat < highlat and lon > lowlon and lon < highlon:
                station_ids.append(row.stid)
        if len(station_ids) == 0:
            if self.verbose:
                print('No stations found within the bounding box.')
                print(bbox)
                print('Try increasing the padding argument.')
            return None
        # download data for each station and save them in a dictionary
        station_dfs = dict()
        for stid in station_ids:
            station_df = self.download_station_data(stid, start_date, end_date)
            if station_df is not None:
                station_dfs[stid] = station_df
        return station_dfs


    def download_shape_object(self,shape,start_date,end_date,padding=1,prj_path=''):
        '''
        Reads a shape file from the specified path and downloads all data for
            all stations within its bounding box

        arguments:
            shape (shapefile.Shape or shapefile.Reader): the shape object to be
                searched for stations
            start_date (datetime.datetime): first date to get data
            end_date (datetime.datetime): last date (inclusive) to get data
            padding (int): an optional parameter that increases or decreases
                the size of the bounding box, must be positive
            prj_path (str): an optional parameter that points to the .prj file
                associated with the shape file; if this specified then the
                bounding box coordinates will be converted to EPSG:4269

        returns:
            dict: a dictionary with station IDs as keys and DataFrames holding
                the downloaded data as values
        '''
        bbox = shape.bbox
        # convert the coordinates to the correct CRS is the path to a
        #   projection file is specified
        if prj_path:
            bbox = self.change_crs(prj_path, bbox)
        # this is a quick check to see if the coordinates are in meters instead
        #   of latitude and longitude, this method isn't perfect but should
        #   catch most issues
        absbox = np.abs(bbox)
        if absbox[0]>180 or absbox[1]>90 or absbox[2]>180 or absbox[3]>90:
            if self.verbose:
                print('It appears that the bounding box {} is not in latitude'
                      ' longitude coordinates. Try specifing the prj_path'
                      ' argument when calling this function.'.format(bbox))
                return None
        df = self.download_bounding_box(bbox,start_date,end_date,padding)
        return df


    def download_shape_file(self,shape_path,start_date,end_date,padding=1,prj_path=''):
        '''
        Reads a shape file from the specified path and downloads all data for
            all stations within its bounding box

        arguments:
            shape_file (str): the path to the shape file to be used
            start_date (datetime.datetime): first date to get data
            end_date (datetime.datetime): last date (inclusive) to get data
            padding (int): an optional parameter that increases or decreases
                the size of the bounding box, must be positive
            prj_path (str): an optional parameter that points to the .prj file
                associated with the shape file; if this specified then the
                bounding box coordinates will be converted to EPSG:4269

        returns:
            dict: a dictionary with station IDs as keys and DataFrames holding
                the downloaded data as values
        '''
        # the whole method is wrapped in a try to make sure we release the
        #   shape file regardless of any exceptions
        try:
            # read the shape file and call the download method for shape objects
            sf = Reader(shape_path)
            df = self.download_shape_object(sf,start_date,end_date,padding,prj_path)
        except FileNotFoundError:
            if self.verbose:
                print('Error: the speficied shape file {} was not found.'.format(shape_path))
            return None
        except Exception as e:
            if self.verbose:
                print('Error: an exception occured attempting to download data'
                      ' for the shape file {}.'.format(shape_path))
                print(e)
                return None
        finally:
            sf.close()
        return df


    def change_crs(self, prj_path, bbox):
        '''
        Changes the CRS of a bounding box to EPSG:4269 used by Mesonet.

        arguments:
            crs_path (str): the path to the .prj file for the original box
            bbox (list): the bounding box to be converted; it expects the same format
                as used in pyshp, [low longitude, low latitude, highlon, highlat]

        returns:
            list: the converted bounding box coordinates
        '''
        # The coordinates for each station in the metadata file come from a
        #   shape file that uses EPSG:4269
        meso_crs = 'EPSG:4269'
        crs_str = ''
        # read the CRS string from the provided file
        try:
            crs_file = open(prj_path)
            crs_str = crs_file.read()
        except FileNotFoundError:
            if self.verbose:
                print('Error: the specified file {} was not found.'.format(prj_path))
                return bbox
        except Exception as e:
            if self.verbose:
                print('Error: unable to read the .prj file {}.'.format(prj_path))
                print(e)
                return bbox
        finally:
            crs_file.close()
        # Create a Transformer object, this is what actually transforms our
        #   coordinates to the proper CRS
        crs = CRS.from_user_input(crs_str)
        transformer = Transformer.from_crs(crs, meso_crs)
        if self.verbose:
            auth = crs.to_authority()
            code = '{}:{}'.format(auth[0],auth[1])
            print('Transforming bounding box from {} to {}.'.format(code,meso_crs))
        # The transformer can only work with points so we have to do each
        #   corner of the box separately
        corner1 = transformer.transform(bbox[0],bbox[1])
        corner2 = transformer.transform(bbox[2],bbox[3])
        return [corner1[1], corner1[0], corner2[1], corner2[0]]



# mesonettools.py
#
# Abhiram Pamula (apamula@okstate.edu)
# Ben Rubinstein (brubinst@hawk.iit.edu)
#
# last updated: 03/20/2023
#
# contains the MesonetTools class
import os, re
import pandas as pd
import numpy as np
import missingno as msno
from .mesonetdownloader import MesonetDownloader
from .etcalculator import ETCalculator
from geopy import distance

class MesonetTools:
    '''
    The MesonetTools class contains methods to assist with processing time
        series generated by the MesonetDownloader class.
    '''

    def __init__(self, verbose=False):
        '''
        init method for the MesonetTools class

        arguments:
            verbose (bool): if true write detailed debugging to stdout
        '''
        self.verbose=verbose
        # these are used internally when replacing data
        self.nondatcols=['STID']
        self.calculatedcols=['TDEW','EVAP']
        # columns that should be summed instead of averaged when resampling
        self.sumcols=['RAIN','EVAP']
        self.errorcodes=[-994,-995,-996,-997,-998,-999]


    def replace_errors(self,df,code=1,column=None):
        '''
        Replace error codes in the dataset with NaN.

        Description of error codes:
            -999 - flagged bad by QA routines
            -998 - sensor not installed
            -997 - missing calibration coefficients
            -996 - station did not report
            -995 - data not reported on this time interval
            -994 - value is too wide to fit in column
        arguments:
            df (DataFrame or dict): the dataframe or dictionary of dataframes
                to be manipulated
            code (int): the specific error code to be replaced, the default 1
                replaces all error codes
            column (str): optional parameter that when specified changes only
                a single column

        returns:
            DataFrame or dict: the modified df object
        '''
        df = df.copy()
        # check if we've been given a dict or dataframe
        if self.__is_dict(df)==-1:
            if self.verbose:
                print('Warning: replace_errors() expects a DataFrame or dict'
                      ' not a {}. No actions performed.'.format(type(df)))
        # check that the error code argument is valid
        elif code != 1 and code not in self.errorcodes:
            if self.verbose:
                print('Warning: {} is not a valid error code. Nothing will'
                      ' be replaced. Use 1 or do not pass in a code argument'
                      ' to replace all error codes or enter one of the'
                      ' following: {}.'.format(code,self.errorcodes))
                print('help(MesonetTools.replace_errors) will give a'
                      ' description of the error codes.')
        # if df is a dictionary, recursively call this function for each of its keys
        elif self.__is_dict(df)==1:
            for key in df:
                df[key] = self.replace_errors(df[key],code,column)
        # if df is a dataframe
        elif self.__is_dict(df)==0:
            if code==1:
                # replace all error codes with NaN
                if column is None:
                    # replace for all columns
                    df = df.replace(self.errorcodes,np.nan)
                else:
                    # check if the column exists
                    if column in df.columns:
                        # replace for a single column
                        df[column] = df[column].replace(self.errorcodes,np.nan)
                    elif self.verbose:
                        print('Warning: there is no column named {}'
                              ' in the dataframe. No actions will be'
                              ' taken.'.format(column))
            # check if code is a valid error code
            else:
                if column is None:
                    # replace for all columns
                    df = df.replace(code,np.nan)
                else:
                    # check if the column exists
                    if column in df.columns:
                        # replace for a single column
                        df[column] = df[column].replace(code,np.nan)
                    elif self.verbose:
                        print('Warning: there is no column named {}'
                              ' in the dataframe. No actions will be'
                              ' taken.'.format(column))
        return df


    def interpolate_missing(self,df,codes=[],column=None):
        '''
        Fills missing data with simple linear interpolation between known
            values
        
        This method will automatically ignore the following columns:
            'STID', 'DATETIME'

        arguments:
            df (DataFrame or dict): the dataframe or dictionary of dataframes
                to be manipulated
            codes (list): optional parameter that when specified interpolates
                only for the specified codes
            column (str): optional parameter that when specified changes only
                a single column

        returns:
            DataFrame or dict: the modified df object
        '''
        # check if we've been given a dict or dataframe
        if self.__is_dict(df)==-1:
            if self.verbose:
                print('Warning: interpolate_missing() expects a DataFrame or'
                      ' dict not a {}. No actions performed.'.format(type(df)))
        # check that at least one error code in the list is valid
        elif codes and all(i not in self.errorcodes for i in codes):
            if self.verbose:
                print('Warning: No valid error codes were entered: {}. No'
                      ' changes will be made. Use an empty list or do not pass'
                      ' in the codes argument to replace all error codes or'
                      ' enter at least one of the following valid error codes:'
                      ' {}'.format(codes,self.errorcodes))
                print('help(MesonetTools.replace_errors) will give a'
                      ' description of the error codes.')
        # if df is a dictionary, recursively call this function for each of its keys
        elif self.__is_dict(df)==1:
            for key in df:
                df[key] = self.interpolate_missing(df[key],codes,column)
        else:
            backup = pd.DataFrame()
            if codes:
                # store a backup so we can recovery error codes that shouldn't
                #   be replaced
                backup = df
            # replace all error codes
            df = self.replace_errors(df,column=column)
            if column is None:
                for ncolumn in df.columns:
                    if ncolumn not in self.nondatcols:
                        df[ncolumn] = df[ncolumn].interpolate()
            else:
                df[column] = df[column].interpolate()
            # if there were specific error codes provided, recover the others
            if codes:
                df = self.__copy_errors(df,backup,codes,column)
        return df


    def fill_neighbor_data(self,df,downloader,codes=[],column=None):
        '''
        Fills missing data with the value from the geographically closest
            station that has the missing observation
        
        This method will automatically ignore -995 error codes and the
            following columns: 'STID', 'DATETIME'

        arguments:
            df (DataFrame or dict): the dataframe or dictionary of dataframes
                to be manipulated
            downloader (MesonetDownloader): a MesonetDownloader object is
                required to calculate distances and download new data
            codes (list): optional parameter that when specified interpolates
                only for the specified codes
            column (str): optional parameter that when specified changes only
                a single column

        returns:
            DataFrame or dict: the modified df object
        '''
        # make sure that downloader is a MesonetDownloader object
        if not isinstance(downloader, MesonetDownloader):
            if self.verbose:
                print('Warning: downloader must be an okmesopy.MesonetDownloader'
                      ' object not a {}. No changes will be made.'.format(type(downloader)))
        # check if we've been given a dict or dataframe
        elif self.__is_dict(df)==-1:
            if self.verbose:
                print('Warning: fill_neighbor_data() expects a DataFrame or'
                      ' dict not a {}. No actions performed.'.format(type(df)))
        # check that at least one error code in the list is valid
        elif codes and all(i not in self.errorcodes for i in codes):
            if self.verbose:
                print('Warning: No valid error codes were entered: {}. No'
                      ' changes will be made. Use an empty list or do not pass'
                      ' in the codes argument to replace all error codes or'
                      ' enter at least one of the following valid error codes:'
                      ' {}'.format(codes,self.errorcodes))
                print('help(MesonetTools.replace_errors) will give a'
                      ' description of the error codes.')
        # if df is a dictionary, recursively call this function for each of its keys
        elif self.__is_dict(df)==1:
            for key in df:
                df[key] = self.fill_neighbor_data(df[key],downloader,codes,column)
        else:
            stid = df.loc[df.index[0],'STID']
            if not codes:
                codes = self.errorcodes.copy()
            # skip -995, no stations will have data on the not sampled intervals
            if -995 in codes: codes.remove(-995)
            for code in codes:
                df = self.replace_errors(df,code,column)
            # create a list of stations sorted by distance
            target_coord = downloader.get_station_coord(stid)
            coord_tuple = list(downloader.metadata.loc[:,['nlat','elon']].itertuples(index = False, name = None))
            for i in coord_tuple:
                if i == target_coord:
                    coord_tuple.remove(i)
            req_loc = downloader.metadata['stid'].loc[downloader.metadata['stid']!=stid]
            station_list=[]
            for i,j in zip(req_loc,coord_tuple):
                station_list.append([i,distance.distance(target_coord,j).miles])
            station_list = sorted(station_list, key = lambda x:(x[1], x[0]))
            stids = [i[0] for i in station_list]
            # we need to deal with nans in calculated columns since no station
            # will have those; first replace them with a placeholder value
            for col in self.calculatedcols:
                if col in df.columns: df[col] = df[col].replace(np.nan,-1000)
            for station in stids:
                # break when all data has been filled
                if df.isnull().sum().sum()==0:
                    break
                df = self.__download_neighbor(df,downloader,station)
            # add the nans back into the calculated columns
            for col in self.calculatedcols:
                if col in df.columns: df[col] = df[col].replace(-1000,np.nan)
        return df


    def summarize_missing(self,df,graph=False):
        '''
        Summarizes error codes in the provided dataframe

        arguments:
            df (DataFrame): the dataframe with missing data to be summarized
            graph (bool): an optional arugument that if true will generate
                visualization of the missing data using the missingno library
        '''
        # Generate a new dataframe that will hold the error info
        cols = df.columns.to_list()
        for col in self.nondatcols: cols.remove(col)
        errors = pd.DataFrame()
        # loop through each column
        for col in cols:
            # count the error codes
            counts = df.groupby(col).size()
            for error in self.errorcodes:
                # populate the error dataframe
                try:
                    count = counts[error]
                    errors.loc[col,error]=count
                except:
                    # a key value error here means there were none of this
                    # error code in this column
                    errors.loc[col,error]=0
            # add a total value for each column
            errors.loc[col,'TOTAL']=errors.loc[col,:].sum()
        # add a total for the error codes as well
        for col in errors.columns:
            errors.loc['TOTAL',col]=errors.loc[:,col].sum()
        # total values, the bottom right cell has the overall total
        total = errors.iloc[-1,-1]
        total_corrected = total - errors.loc[errors.index[-1],-995]
        # print a summary of the missing data
        print('Missing data summary for {} station:'.format(df.loc[df.index[0],'STID']))
        print('-----------------------')
        print('To see a description of each error code run help(MesonetTools.replace_errors)')
        print('Note: the -995 error code is used when data is not collected'
              ' on an interval. This code is generally normal and expected.')
        print('There are {} total missing data points and {} missing data'
              ' data points excluding -995 codes'.format(int(total),int(total_corrected)))
        print('The following chart displays the number of each kind of error'
              ' code found in each column of the DataFrame.\n')
        print(errors.astype(int))
        # graphically display errors if the chart arguement is true
        if graph:
            # prepare a dataframe for the missingno library, we need to
            # actually replace all the error codes with NaN so missingno can
            # recognize them
            rep_df = self.replace_errors(df)
            # also remove non data columns
            for col in self.nondatcols:
                rep_df.drop(col,axis=1,inplace=True)
            msno.matrix(rep_df,freq='5M')


    def save_timeseries(self,df,column,step=5,get_min=False,get_max=False,
                        timezone=None):
        '''
        Saves the a data for a single variable for a single station as a
            PyHSPF readable timeseries. If a set of stations is provided,
            the arithmetic mean is used instead

        arguments:
            df (DataFrame or dict): the DataFrame to create a series from; if
                a dictionary is passed in, the arithmetic mean of all contained
                DataFrames is used
            column (str): the variable to create a time series for
            step (int or str): the time interval to use in minutes; must be at
                least 5 and divisible by 5; the closest multiple of 5 will be
                least instead if not; the strings 'hourly' and 'daily' are also
                valid
            get_min (bool): If true, returns the minimum value instead of mean
                or sum when resampling
            get_max (bool): If true, returns the maximum value instead of mean
                or sum when resampling
            timezone (str): 

        returns:
            tuple (int,datetime,list): the timeseries object in the form
                (step size,start date,data)
        '''
        column = column.upper()
        # validate the step size
        if step == 'hourly': step = 60
        elif step == 'daily': step = 1440
        elif step < 5:
            if self.verbose:
                print('{} was given as step size but it must be at least 5'
                      ' minutes. A step size of 5 minutes will be used'
                      ' instead'.format(step))
            step = 5
        elif step%5 != 0:
            if self.verbose:
                print('Step size must be a multiple of 5. The given step size'
                      ', {}, will be rounded up to {}.'.format(step,step + 5 - step%5))
            step = step + 5 - step%5
        # if a dictionary has been given, generate the average
        tempdf = pd.DataFrame()
        if self.__is_dict(df):
            # concatenate all the DataFrames together, using only the relevant
            #   column to save memory
            for key in df:
                start = df[key].index[0]
                # make sure the column exists
                if not column in df[key].columns:
                    print('Error: the column {} was not found in the dataframe'
                          ' for the {} station.'.format(column,key))
                    return None
                tempdf = pd.concat((tempdf,df[key].copy(column)))
            # remove all error codes so they don't throw off the averages
            tempdf = self.replace_errors(tempdf)
            # make df the mean
            df = tempdf.groupby(tempdf.index).mean(numeric_only=True)
        else:
            # get the start date and time
            start = df.index[0]
        # easiest case we just use every data point
        if step == 5:
            data = df[column].replace(self.errorcodes,np.nan).tolist()
        else:
            # check what kind of resampling we need to do
            if get_min and not get_max:
                data = df[column].replace(self.errorcodes,np.nan).resample(str(step)+'Min').min().tolist()
            elif get_max and not get_min:
                data = df[column].replace(self.errorcodes,np.nan).resample(str(step)+'Min').max().tolist()
            elif get_min and get_max:
                print('Error: both get_min and get_max were set to True. Only'
                      ' one can be true.')
                return 
            elif column in self.sumcols:
                data = df[column].replace(self.errorcodes,np.nan).resample(str(step)+'Min').sum().tolist()
            else:
                data = df[column].replace(self.errorcodes,np.nan).resample(str(step)+'Min').mean().tolist()
        ts = step, start, data
        return ts


    def save_csv(self,df,path,force=False,filename=''):
        '''
        Saves a downloaded dataset as a CSV file.

        arguments:
            df (DataFrame or dict): the DataFrame or dictionary of DataFrames
                to be saved
            path (str): the directory to save the file to
            force (bool): if set to true will overwrite existing files
            filename (str): optional name of the csv file; filenames must be
                alphanumeric characters, hyphens, underscores, and periods only
        '''
        path = os.path.normpath(path)
        try:
            if not os.path.isdir(path): os.mkdir(path)
        except PermissionError:
            errstr = ('The specified directory {} does not exist and you do not'
                      ' have permission to create it. Please try a different'
                      ' directory.')
            raise PermissionError(errstr.format(path))
        # check if the filename is valid
        if re.search(r'[^A-Za-z0-9\._\-\\]',filename):
            print('Warning: Filenames can only contain alphanumeric characters'
                  ', hyphens, underscores, and periods. Enter a new filename'
                  ' or leave the argument blank to use an autogenerated name.')
            return
        # if we have a dictionary, concatenate all the DataFrames together
        if self.__is_dict(df):
            concat_df = pd.DataFrame()
            for key in df:
                concat_df = pd.concat((concat_df,df[key]))
                # generate the file name
                if not filename:
                    start = df[key].index[0].strftime('%m%d%y')
                    end = df[key].index[-1].strftime('%m%d%y')
                    filename = '{}_and-{}-more_{}-{}.csv'.format(key,len(df)-1,start,end)
            # if file exists check if the force argument is true
            if os.path.exists('{}/{}'.format(path,filename)) and not force:
                print('Warning: The file {} already exists in the directory {}.'
                      ' Please choose a new filename or directory or set the force'
                      ' argument to true to overwrite the existing file.'.format(filename,path))
                return
            concat_df.to_csv('{}/{}'.format(path,filename))
        else:
            if not filename:
                stid = df.loc[df.index[0],'STID']
                start = df.index[0].strftime('%m%d%y')
                end = df.index[-1].strftime('%m%d%y')
                filename = '{}_{}-{}.csv'.format(stid,start,end)
            # if file exists check if the force argument is true
            if os.path.exists('{}/{}'.format(path,filename)) and not force:
                print('Warning: The file {} already exists in the directory {}.'
                      ' Please choose a new filename or directory or set the force'
                      ' argument to true to overwrite the existing file.'.format(filename,path))
                return
            df.to_csv('{}/{}'.format(path,filename))


    def calculate_dewpoint(self, df, a=17.625, b=243.04):
        '''
        Calculates the dewpoint temperatures from temperature and relative
        humidity using the Magnus-Tetens formula
        DOI:10.1175/BAMS-86-2-225
        
        arguments:
            df (DataFrame or dict): the dataframe or dictionary of dataframes
                to calculate dewpoints for
            a, b (float): the Magnus coefficients, by default the ones given by
                Alduchov and Eskridge are used:
                DOI:10.1175/1520-0450(1996)035<0601:IMFAOS>2.0.CO;2
        '''
        # make sure we have been given a dataframe or dictionary
        if self.__is_dict(df)==-1:
            if self.verbose:
                print('Warning: calculate_dewpoint() expects a DataFrame or'
                      ' dict not a {}. No actions performed.'.format(type(df)))
        # if df is a dictionary, recursively call this function for each of its keys
        elif self.__is_dict(df)==1:
            for key in df:
                df[key] = self.calculate_dewpoint(df[key],a,b)
        # if df is a dataframe we can find the dewpoint
        elif self.__is_dict(df)==0:
            # replace errors in RELH and TAIR to ensure we get nans where data
            # is missing
            rep_df = self.replace_errors(df,column='TAIR')
            rep_df = self.replace_errors(rep_df,column='RELH')
            alpha = np.log(rep_df["RELH"]/100) + (a*rep_df["TAIR"]/(b+rep_df["TAIR"]))
            df["TDEW"] = (b * alpha) / (a - alpha)
        return df


    def calculate_ret(self,df,start,end,downloader,timestep=60,wind='WS2M',
                      error_handling='nan',timezone='US/Central'):
        '''
        Calculates the reference evapotranspiration using the Penman-Monteith
            equation. This method sets everything up using a MesonetDownloader
            dataframe and the actual calculations are done in the ETCalculator
            class.
        
        arguments:
            df (DataFrame): the dataframe to calculate reference ET for
            start (datetime.datetime): first date to get data
            end (datetime.datetime): last date (inclusive) to get data
            downloader (MesonetDownloader): a MesonetDownloader object is
                required to get the station location data
            timestep (int): time step to use in minutes; acceptable values are
                60, 1440, or a factor of 60 that is also a multiple of 5
            wind (str): which column to use for wind speed data; acceptable
                values are 'WS2M','WSPD', and 'WMAX' which are wind speed at 2
                meters, 10 meters, and max wind speed at 10 meters respectively
                it is recommended to use the default 'WS2M' so that no
                correction needs to be done
            error_handling (str): tells the function how to deal with error
                codes in the data; acceptable values are 'nan', 'interpolate',
                and 'neighbor'. 'nan' is the default and simply replaces error
                codes with nan. This will result in nans in the ET time series.
                'interpolate' and 'neighbor' use the interpolate_missing and
                fill_neighbor_data methods respectively to fill in missing data

        returns:
            tuple (int,datetime,list): step size in minutes, start time
                of the series, reference evapotranspiration in mm
        '''
        # validate the timestep
        if timestep == 'daily' or timestep == 1440: timestep = 1440
        elif timestep == 'hourly' or timestep == 60: timestep = 60
        elif timestep < 5:
            if self.verbose:
                print('{} was given as step size but it must be at least 5'
                      ' minutes. A step size of 5 minutes will be used'
                      ' instead'.format(timestep))
            timestep = 5
        elif timestep%5 != 0 and 60%timestep != 0:
            if self.verbose:
                msg = ('Error: step size was given as {}. Valid time steps are'
                       ' hourly, daily, or a factor of 60 that is also a'
                       ' multiple of 5.')
                raise ValueError(msg)
        # prepare the dataframe
        if error_handling == 'nan':
            df = self.replace_errors(df)
        elif error_handling == 'interpolate':
            df = self.interpolate_missing(df,column='TAIR')
            df = self.interpolate_missing(df,column='RELH')
            df = self.interpolate_missing(df,column=wind)
            df = self.interpolate_missing(df,column='SRAD')
        elif error_handling == 'neighbor':
            df = self.fill_neighbor_data(df, downloader,column='TAIR')
            df = self.fill_neighbor_data(df, downloader,column='RELH')
            df = self.fill_neighbor_data(df, downloader,column=wind)
            df = self.fill_neighbor_data(df, downloader,column='SRAD')
        else:
            print('Invalid error code handling method was given. Acceptable'
                  ' values for the error_handling parameter are \'nan\','
                  ' \'interpolate\', and \'neighbor\'.')
            return None
        df = self.calculate_dewpoint(df)
        # convert the index to local time
        df = df.tz_localize('UTC').tz_convert(timezone)
        # initialize the ETCalculator object
        calc = ETCalculator(timestep=timestep)
        # add location data to the calculator
        stid = df.iloc[0,0]
        calc.add_location(*downloader.get_location(stid))
        # add the required time series to the calculator
        if timestep == 1440:
            # daily calculations use min/max temperatures insetad of average
            calc.add_timeseries('tmin',*self.save_timeseries(df,'TAIR',timestep,get_min=True))
            calc.add_timeseries('tmax',*self.save_timeseries(df,'TAIR',timestep,get_max=True))
            calc.add_timeseries('dewpoint',*self.save_timeseries(df,'TDEW',timestep))
            calc.add_timeseries('wind',*self.save_timeseries(df,wind,timestep))
            calc.add_timeseries('solar',*self.save_timeseries(df,'SRAD',timestep))
        else:
            # hourly and fractional hourly calculations are done the same way
            calc.add_timeseries('temperature',*self.save_timeseries(df,'TAIR',timestep))
            calc.add_timeseries('dewpoint',*self.save_timeseries(df,'TDEW',timestep))
            calc.add_timeseries('wind',*self.save_timeseries(df,wind,timestep))
            calc.add_timeseries('solar',*self.save_timeseries(df,'SRAD',timestep))
        # calculate RET and return the time series
        calc.calculate_ret(start,end)
        return (timestep, *calc.data['RET'])


    def __download_neighbor(self,df,downloader,station_id):
        '''
        Helper function that downloads and fills data from a neighboring station

        arguments:
            df (DataFrame): the dataframe with missing data to be filled 
            downloader (MesonetDownloader): a MesonetDownloader object is
                required to download new data
            station_id (str): the station ID for the neighboring station

        returns:
            DataFrame: the modified df object
        '''
        # get a list of dates with missing data
        missing_dates = []
        for dt in df[df.isna().any(axis=1)].index.unique():
            if dt.date() not in missing_dates: missing_dates.append(dt.date())
        # download data for each of the missing dates
        for miss_date in missing_dates:
            date = pd.to_datetime(miss_date)
            neighbor_df = downloader.download_station_data(station_id,date,date)
            if neighbor_df is not None:
                # we don't want to copy over any error codes so replace all of them
                neighbor_df = self.replace_errors(neighbor_df)
                # fill in data
                df = df.fillna(neighbor_df)
        return df


    def __copy_errors(self,df,backup,codes,column=None):
        '''
        Helper function that copies error codes back into a dataframe

        arguments:
            df (DataFrame): the dataframe to be manipulated
            backup (DataFrame): a copy of df with error codes still in place 
            codes (list): a list of codes to copy back into df
            column (str): optional parameter that when specified changes only
                a single column

        returns:
            DataFrame: the modified df object
        '''
        # TODO: fix the SettingWithCopyError?
        pd.options.mode.chained_assignment = None
        for code in self.errorcodes:
            if code not in codes:
                if column is not None:
                    # these commands cause a SettingWithCopyError. I think it
                    #   is a false positive but I'm not fully sure I
                    #   understand the error properly
                    df[column].loc[backup[column]==code] = code
                else:
                    for ncolumn in df.columns:
                        df[ncolumn].loc[backup[ncolumn]==code] = code
        pd.options.mode.chained_assignment = 'warn'
        return df


    def __is_dict(self,df):
        '''
        MesonetDownloader creates single dataframes and dictionaries of
            dataframes. Returns 1 for dict, 0 for dataframe, and -1 as an
            error code for anything else

        arguments:
            df (DataFrame or dict): the object to type check

        returns:
            int: 1 for a dictionary, 0 for a DataFrame, -1 otherwise
        '''
        if isinstance(df,dict):
            return 1
        elif isinstance(df,pd.DataFrame):
            return 0
        else:
            return -1



# etcalculator.py
#
# Abhiram Pamula (apamula@okstate.edu)
# Ben Rubinstein (brubinst@hawk.iit.edu)
#
# adapted from PyHSPF by David Lampert (https://github.com/djlampert/PyHSPF)
#
# last updated: 03/20/2023
#
# contains methods to calculate reference evapotranspiration

import datetime
import numpy as np

class ETCalculator:
    '''
    The ETCalculator class contains methods to calculate reference
        evapotranspiration using the Penman-Monteith Equation
        https://www.fao.org/3/x0490e/x0490e06.htm
    '''

    def __init__(self, verbose=False, timestep=60):
        '''
        Init method for ETCalculator class
        
        arguments:
            verbose (bool): if true write detailed debugging to stdout
            timestep (int or str): the time interval to use in minutes; must be
                hourly, daily, or a multiple of 5 AND factor of 60. Hourly or
                daily timesteps can be entered as 'hourly' or 60 or 'daily' or
                1440
        '''
        self.verbose = verbose

        # validate the step size
        if timestep == 'daily' or timestep == 1440: timestep = 1440
        elif timestep == 'hourly' or timestep == 60: timestep = 60
        elif timestep < 5:
            if self.verbose:
                print('{} was given as step size but it must be at least 5'
                      ' minutes. A step size of 5 minutes will be used'
                      ' instead'.format(timestep))
            timestep = 5
        elif timestep%5 != 0 and 60%timestep != 0:
            if self.verbose:
                msg = ('Error: step size was given as {}. Valid time steps are'
                       ' hourly, daily, or a factor of 60 that is also a'
                       ' multiple of 5.')
                raise ValueError(msg)
        self.timestep = timestep

        # Dictionary to hold required climate data, varies depending of if we
        # are working with hourly/fractional hourly or daily data
        if timestep == 1440:
            self.data = {
                'tmin':        None,
                'tmax':        None,
                'dewpoint':    None,
                'wind':        None,
                'solar':       None,
                }
        else:
            self.data = {
                'temperature': None,
                'dewpoint':    None,
                'wind':        None,
                'solar':       None,
                }

        # location
        self.longitude = None
        self.latitude  = None
        self.elevation = None


    def add_location(self, latitude, longitude, elevation):
        '''
        Adds location information, required for some calculations
        
        arguments:
            longitude (float): longitude in degrees
            latitude (float): latitude in degrees
            elevation (float): elevation in meters
        '''

        self.longitude = longitude
        self.latitude  = latitude
        self.elevation = elevation


    def add_timeseries(self, tstype, timestep, start, values):
        '''
        Adds a timeseries to use for ET calculations
        
        arguments:
            tstype (str): the type of data being given, must match one of the
                keys in the self.data dictionary
                (temperature, dewpoint, humidity, wind, solar, RET)
            timestep (int or str): the interval in minutes between datapoints,
                must match the one given when initializing ETCalculator
            start (datetime): the datetime for the first datapoint in the
                timeseries
            values (list): a list of values for the measured data
        '''
        if timestep == 'daily': timestep = 1440
        elif timestep == 'hourly': timestep = 60
        # validate the timestep
        if not timestep == self.timestep:
            print('Warning: the timestep for this series ({}) does not match'
                  ' the timestep of this calculator ({}). The timeseries will'
                  ' not be added.'.format(timestep,self.timestep))
            return

        # validate the type of timeseries
        if tstype not in self.data:
            print('Warning: unknown time series type {} specified'.format(tstype))
            print('valid choices are ' + ', '.join(self.data) + '. The'
                  ' timeseries will not be added.')
            return

        self.data[tstype] = start, values


    def rad(self, d): 
        '''
        Converts degrees to radians

        arguments:
            d (float): angle in degrees

        returns:
            float: angle in radians
        '''

        return (np.pi / 180 * d)


    def atmosphere_pressure(self):
        '''
        Estimates the atmospheric pressure (kPa) from the elevation (m) using a 
            simplified ideal gas law.

        returns:
            float: estimate atmospheric pressure in kPa
        '''

        if self.elevation is None:
            raise RuntimeError('Elevation not specified. Use add_location()'
                               ' to define the location.')

        return 101.3 * ((293.0 - 0.0065 * self.elevation) / 293.0)**5.26


    def vapor_pressure(self, temp):
        '''
        Estimates the water vapor pressure (kPa) at a given temperature (C).

        arguments:
            T (float): temperature in C

        returns:
            float: water vapor pressure in kPa
        '''

        return 0.6108 * np.exp((17.27 * temp) / (temp + 237.3))


    def wind_correction(self, u1, z = 10):
        '''
        Estimates the wind speed at z = 2 m from the wind speed at height z 
            Mesonet records wind speed at both 2 and 10 meters, this allows the
            user to choose which to use

        arguments:
            u1 (float): wind speed at z meters in m/s
            z (float): height of the measurement in meters

        returns:
            float: estimated wind speed at 2 meters
        '''

        # note that at z = 2 the multiplier is 1.000 ~ 1
        return u1 * 4.87 / np.log(67.8 * z - 5.42)


    def get_soil(self, times, Gday, Gnight):
        '''
        Provides an array of the values of the soil heat flux if a list is 
        provided, otherwise just provides the appropriate day or night value.

        arguments:
            times (list): list of datetime.datetime instances
            Gday (float): soil heat flux during the day in MJ/m2/hour
            Gnight (float): soil heat flux during the night in MJ/m2/hour

        returns:
            list(float): list of heat flux values
        '''

        Gd = np.array([Gday if t else Gnight 
                          for t in self.is_daytime(times)])

        return Gd


    def get_Cd(self, times, Cday, Cnight):
        '''
        Provides an array of the values denominator coefficient if a list is 
        provided, otherwise just provides the appropriate day or night value.

        arguments:
            times (list): list of datetime.datetime instances
            Cday (float): daytime denominator coefficient
            Cnight (float): nighttime denominator coefficient

        returns:
            list(float): list of heat flux values
        '''

        Cd = np.array([Cday if t else Cnight 
                          for t in self.is_daytime(times)])
        return Cd


    def time_round(self, d):
        """Rounds a datetime.datetime instance to the nearest hour."""

        yr, mo, da, hr, mi = d.timetuple()[:5]

        if 30 <= d.minute:  # round to the next hour

            r = (datetime.datetime(yr, mo, da, hr) + 
                 datetime.timedelta(hours = 1))

        else:

            r = datetime.datetime(yr, mo, da, hr)

        return r


    def sun(self, time, zenith = 90.83, nearest_hour = False):
        '''
        Estimates the sunrise and sunset time for a given latitude, longitude,
            and date and optionally rounds it to the nearest hour. Algorithm
            is derived from "Almanac for Computers," 1990, Nautical Almanac,
            Office United States Naval Observatory, Washington DC 20392.

        arguments:
            time (datetime): the date to check sunrise and sunset for
            zenith (float): sun's zenith distance at a certain time, 90°50' for
                sunrise/sunset (see p. B6 of the reference)
            nearest_hour (bool): whether or not to round the times to the
                nearest hour

        returns:
            tuple (datetime): a pair of datetime.datetime objects in the form
                (sunrise, sunset)
        '''

        # calculate day of the year

        N = time.timetuple().tm_yday

        # convert longitude to hour value and calculate approximate time

        hour = self.longitude / 15

        # rise

        t = N + (6 - hour) / 24

        # calculate mean anomaly

        M = (360 / 365.25 * t) - 3.289

        # calculate sun's true longitude

        L = (M + 1.916 * np.sin(self.rad(M)) + 
             (0.02 * np.sin(2 * self.rad(M))) + 282.634)

        # adjust into 0, 360 range

        if L < 0:   L+=360
        if L > 360: L-=360
             
        # calculate right ascension

        RA = 180 / np.pi * np.arctan(0.91764 * np.tan(self.rad(L)))

        # ensure RA is in the same quadrant as L

        Lquadrant  = (np.floor(L  / 90)) * 90
        RAquadrant = (np.floor(RA / 90)) * 90

        RA = RA + (Lquadrant - RAquadrant)

        # convert to hours

        RA = RA / 15

        # calculate the solar declination

        sinDec = 0.39782 * np.sin(self.rad(L))
        cosDec = np.cos(np.arcsin(sinDec))

        # calculate the solar local hour angle

        cosH = ((np.cos(self.rad(zenith)) - sinDec * 
                 np.sin(self.rad(self.latitude))) / 
                (cosDec * np.cos(self.rad(self.latitude))))

        # rise

        H = 360 - 180 / np.pi * np.arccos(cosH)

        H = H / 15

        # calculate local mean time of rising

        T = H + RA - 0.06571 * t - 6.622

        # adjust back to UTC

        UT = T - hour

        # adjust to local time zone (this is not super accurate, but for the 
        # purposes of ET estimation really doesn't warrant detailed estimation)

        offset = round(self.longitude / 15)

        # daily saving time

        if time.month < 3 or 10 < time.month: offset -= 1

        localT = UT + offset

        # adjust to [0, 24] as needed

        if localT < 0:  localT += 24
        if localT > 24: localT -= 24
            
        t = datetime.timedelta(hours = localT)

        sunrise = ':'.join(str(t).split(':')[:2])

        sunrise = datetime.datetime(*time.timetuple()[:3]) + t

        # now do sunset

        t = N + (18 - hour) / 24

        # calculate mean anomaly

        M = (360 / 365.25 * t) - 3.289

        # calculate sun's true longitude

        L = (M + 1.916 * np.sin(self.rad(M)) + 
             (0.02 * np.sin(2 * self.rad(M))) + 282.634)

        # adjust into 0, 360 range

        if L < 0:   L+=360
        if L > 360: L-=360
             
        # calculate right ascension

        RA = 180 / np.pi * np.arctan(0.91764 * np.tan(self.rad(L)))

        # ensure RA is in the same quadrant as L

        Lquadrant  = (np.floor(L  / 90)) * 90
        RAquadrant = (np.floor(RA / 90)) * 90

        RA = RA + (Lquadrant - RAquadrant)

        # convert to hours

        RA = RA / 15

        # calculate the solar declination

        sinDec = 0.39782 * np.sin(self.rad(L))
        cosDec = np.cos(np.arcsin(sinDec))

        # calculate the solar local hour angle

        cosH = ((np.cos(self.rad(zenith)) - sinDec * 
                 np.sin(self.rad(self.latitude))) / 
                (cosDec * np.cos(self.rad(self.latitude))))

        # set

        H = 180 / np.pi * np.arccos(cosH) / 15

        # calculate local mean time of rising

        T = H + RA - 0.06571 * t - 6.622

        # adjust back to UTC

        UT = T - hour

        # adjust to local time zone

        localT = UT + offset

        # adjust to [0, 24] as needed

        if localT < 0:  localT += 24
        if localT > 24: localT -= 24
            
        t = datetime.timedelta(hours = localT)

        sunset = ':'.join(str(t).split(':')[:2])

        sunset = datetime.datetime(*time.timetuple()[:3]) + t

        if nearest_hour: 
            
            sunrise, sunset = self.time_round(sunrise), self.time_round(sunset)

        return sunrise, sunset


    def is_daytime(self, times):
        '''
        Determines whether the sun is up or not for each values in "times"

        arguments:
            times (list): list of datetime.datetime instances to check

        returns:
            list (bool): a list of boolean values where true means the sun is
                up
        '''

        sunrise, sunset = zip(*[self.sun(t) for t in times])

        return [r <= t and t < s for t, r, s in zip(times, sunrise, sunset)]


    def daily_radiation(self, times, solar, Tmin, Tmax, Pv, albedo = 0.23,
                        sigma = 4.903e-9,):
        '''
        Estimates the net radiation (MJ/m2/day) from the measured downwelling
            radiation, min/max temperature, and vapor pressure

        arguments:
            times (list): list of datetime.datetime instances
            solar (list): measured solar radiation in MJ/m2/hour (as floats);
                R_s in the ASCE document
            Tmin (float): daily minimium temperature in K
            Tmax (float): daily maximum temperature in K
            Pv (float): vapor pressure in kPa
            albedo (float): assumed value typical for grass
            sigma (float): Boltzmann constant in MJ/m2/day

        returns:
            list (float): time series of net solar radiation
        '''

        # make sure the location has been supplied

        location = self.longitude, self.latitude, self.elevation

        if any([l is None for l in location]):
            print('error: location has not been specified\n')
            raise

        # convert the dates to julian days

        julian = np.array([t.timetuple().tm_yday for t in times])

        # calculate the solar declination (rad)

        sd = 23.45 * self.rad(np.cos(2 * np.pi / 365 * (julian - 172)))

        # calculate the inverse relative distance Earth-Sun

        irl = 1 + 0.033 * np.cos(2 * np.pi / 365 * julian)

        # calculate the hour angle at sunset (rad)

        sha = np.arccos(-np.tan(self.rad(self.latitude)) *
                           np.tan(sd))

        # calculate the extraterrestrial radiation

        et_rad = 37.59 * irl * (sha *
                                np.sin(self.rad(self.latitude))
                                * np.sin(sd) +
                                np.cos(self.rad(self.latitude)) *
                                np.cos(sd) * np.sin(sha))

        # calculate the clear day solar radiation

        clear = (0.00002 * self.elevation + 0.75) * et_rad

        # shortwave radiation

        Rns = solar * (1 - albedo)

        # longwave radiation

        Rnl = (sigma * 0.5 * (Tmin**4 + Tmax**4) *
               (0.34 - 0.139 * np.sqrt(Pv)) *
               (1.35 * solar / clear - 0.35))

        return Rns - Rnl


    def hourly_radiation(self, times, solar, T, Pv, albedo = 0.23,
                         sigma = 2.042e-10, Gsc = 4.92):
        '''
        Estimates the net radiation (MJ/m2/hour) from the measured downwelling
            radiation, temperature, and vapor pressure

        arguments:
            times (list): list of datetime.datetime instances
            solar (list): measured solar radiation in MJ/m2/hour (as floats);
                R_s in the ASCE document
            T (float): average temperature in K
            Pv (float): vapor pressure in kPa
            albedo (float): assumed value typical for grass
            sigma (float): Boltzmann constant in MJ/m2/hour
            Gsc (float): solar constant in MJ/m2/hour

        returns:
            list (float): time series of net solar radiation
        '''

        # time step in hours
        t1 = self.timestep / 60

        # convert dates to julian days

        julian  = np.array([t.timetuple().tm_yday for t in times])
        ts      = np.array([t.timetuple().tm_hour for t in times])
        tm      = np.array([t.timetuple().tm_min for t in times])

        # calculate the solar declination (rad)

        d = self.rad(0.409 * np.cos(2 * np.pi / 365 * julian - 1.39))

        # calculate the inverse relative distance factor

        dr = 1 + 0.033 * np.cos(2 * np.pi / 365 * julian)

        # calculate the seasonal correction for solar time

        b  = 2 * np.pi * (julian - 81) / 364
        Sc = (0.1645 * np.sin(2 * b) - 0.1255 * np.cos(b) - 
              0.025 * np.sin(b))

        # convert the latitude to radians

        p = self.rad(self.latitude)

        # calculate the longitude at the center of the time zone

        Lm = self.longitude
        Lz = 15 * round(self.longitude / 15)

        # calculate the sunset hour angle

        ws = np.arccos(-1 * np.tan(p) * np.tan(d))

        # calculate the solar time angle at the midpoint of the time interval

        w = np.pi / 12 * ((ts + tm/60 + t1/2 + (Lm - Lz) / 15 + Sc) - 12)

        # calculate the solar time angles

        w1 = w - np.pi * t1 / 24
        w2 = w + np.pi * t1 / 24
        # overwrite the values if the sun is down

        w1[w1 < -ws] = -ws[w1 < -ws]
        w2[w2 < -ws] = -ws[w2 < -ws]
        w1[w2 < -ws] =  ws[w2 < -ws]
        w2[w2 >  ws] =  ws[w2 >  ws]
        w1[w1 >  w2] =  w2[w1 >  w2]

        # calculate the extraterrestrial radiation for the hour

        c = 12 / np.pi * Gsc * dr
        Ra = c * ((w2 - w1) * np.sin(p) * np.sin(d) +
                  np.cos(p) * np.cos(d) * (np.sin(w2) - np.sin(w1)))

        # calculate the clear day solar radiation

        Rso = (0.00002 * self.elevation + 0.75) * Ra

        # calculate the angle of the sun above the horizon at the midpoint
        # of the hour period

        B = np.arcsin(np.sin(p) * np.sin(d) + 
                         np.cos(p) * np.cos(d) * np.cos(w))

        # calculate ratio of the clear sky radiation to the actual radiation

        fcd = np.array([1.35 * (s / c) - 0.35 
                           if 0 < c and 0.3 < s / c and 17 < a 
                           else 0.05
                           for s, c, a in zip(solar, Rso, B)])

        Rns = solar * (1 - albedo)
        Rnl = sigma * 0.5 * T**4 * (0.34 - 0.139 * np.sqrt(Pv)) * fcd

        return Rns - Rnl


    def calculate_ret(self, start, end, albedo = 0.23, Cn = None, Cday = 0.24,
                      Cnight = 0.96, Cd = 0.34, Gday = 0.1, Gnight = 0.5,
                      Gd = 0, wheight = 2):
        '''
        Calculates the potential evapotransporation (PET) in mm for an hourly
        timeseries using the Penman-Monteith equation. Equations from:
        
        American Society of Civil Engineers (ASCE)
        Task Committee on Standardization of Reference Evapotranspiration
        Environmental and Water Resources Institute
        THE ASCE Standardized Reference Evapotranspiration Equation
        January 2005 Final Report

        https://xwww.mesonet.org/images/site/ASCE_Evapotranspiration_Formula.pdf

        arguments:
        start (datetime): the first day to calculate RET for
        end (datetime): the last day, exclusive
        albedo (float): i.e. canopy coefficient; ASCE recommends using a fixed
            value of 0.23
        Cn (float): numerator coefficient, varies depending on short or tall
            crop reference and hourly or daily time step; if not specified the
            correct Cn for short reference will be chosen based on timestep
            see Table 1 in the ASCE document
        Cday (float): daytime denominator coefficient; used for non-daily
            calculations, by default the short crop reference value is used
        Cnight (float): nighttime denominator coefficient
        Cd (float): daily denominator coefficient; by default the short crop
            reference value is used
        Gday (float): ground heat flux coefficient during the day; by default
            the short crop reference value is used; see eqns. 65a - 66b
        Gnight (float): ground heat coefficient flux during the night
        Gd (float): daily soil heat flux density; 0 by default, see eqn. 30
        wheight (float): the actual height wind measurements were taken at;
            used to correct the wind speed to 2m; should be 2 if WS2M is used
            and 10 if WSPD or WMAX is used
        '''

        if self.verbose:
            print('Calculating reference evapotranspiration...\n')

        # check the location has been supplied

        location = self.longitude, self.latitude, self.elevation

        if any([v is None for v in location]):
            print('Warning: location must be specified. Use the add_location()'
                  ' method.')
            return
        
        # check that all the time series are present

        required = [*self.data]

        if any([self.data[ts] is None for ts in required]):
            for ts in required:
                if self.data[ts] is None:
                    print('Warning: {} data unavailable. Add the required data'
                          ' with the add_timeseries() method.'.format(ts))
            return

        # make a list of times to use to calculate values

        times = [start + datetime.timedelta(minutes = self.timestep) * i 
                 for i in range(int((end-start).days * 1440/self.timestep))]

        # create a dictionary of numpy arrays for the requested period

        data = {}

        for ts in required:

            s, d = self.data[ts]

            # find the index of the requested start and end times

            i = int((start - s).days * 1440 / self.timestep)
            j = int((end - s).days * 1440 / self.timestep)

            # make the time series (if the data are available)

            try:
                data[ts] = np.array(d[i:j])
            except:
                print('Warning: {} data unavailable '.format(ts) +
                      'for requested period {} -- {}\n'.format(start, end))
                return

        # check and replace dewpoints higher than tmin/temperature
        #   (physically impossible)
        # convert solar from watt/m2 to MJ/m2/(hour or day)
        # calculate average temp for daily time step
        if self.timestep == 1440:
            dewpoint = np.minimum(data['tmin'], data['dewpoint'])
            solar = data['solar'] * 86400 / 10**6
            data['temperature'] = (data['tmin']+data['tmax'])/2
        else:
            dewpoint = np.minimum(data['temperature'], data['dewpoint'])
            solar = data['solar'] * 3600 / 10**6

        # set the numerator coefficient if it hasn't been provided
        if Cn is None:
            if self.timestep == 1440: Cn = 900
            else: Cn = 37

        # correct windspeed height (note the correction is essentially 1 if z
        # is 2 meters)

        u2 = self.wind_correction(data['wind'], z = wheight)

        # estimate the atmospheric pressure at the given elevation

        P = self.atmosphere_pressure()

        # estimate the average saturation vapor pressure (kPa)

        Ps = self.vapor_pressure(data['temperature'])

        # estimate the vapor pressure (kPa) from the dew point

        Pv = self.vapor_pressure(dewpoint)

        # estimate the vapor pressure curve slope (kPa C-1) at mean temperature

        d = (4098 * self.vapor_pressure(data['temperature']) / 
             (data['temperature'] + 237.3)**2)

        # convert C to K

        T = data['temperature'] + 273.15

        # adjust values based on the timestep
        if self.timestep == 1440:
            # daily soil heat flux is 0 by default
            soil = Gd
            # C_d coefficient is constant for daily estimation
            C_d = Cd
            Tmin = data['tmin'] + 273.15
            Tmax = data['tmax'] + 273.15
            rnet = self.daily_radiation(times, solar, Tmin, Tmax, Pv)
        else:
            rnet = self.hourly_radiation(times, solar, T, Pv)
            # hourly soil heat flux varies based on night/day
            soil = rnet * self.get_soil(times, Gday, Gnight)
            # C_d varies depending on night/day
            C_d = self.get_Cd(times, Cday, Cnight)
        
        # estimate the psychrometric constant (kPa C-1)
        # equation is gamma = cp(T) * P * MWair / latent_heat / MWwater
        # this uses cp for water at T = 20 C

        g = 0.000665 * P

        # estimate the reference evapotranspiration
        if self.timestep == 1140:
            RET = (0.408*d*(rnet-soil)+g*Cn/T*u2*(Ps-Pv))/(d+g*(1+Cd*u2))
        else:
            RET = ((0.408 * (rnet - soil) * d  + Cn * u2 / T * (Ps - Pv) * g) / 
                  (d + (g * (1 + C_d * u2)))) * (60 / self.timestep) 

        self.data['RET'] = start, RET

        if self.verbose: 

            print('Finished calculating reference evapotranspiration\n')



'''
init file
'''

__all__ = ['mesonetdownloader','mesonettools','etcalculator']

from .mesonetdownloader     import MesonetDownloader
from .mesonettools          import MesonetTools
from .etcalculator          import ETCalculator



from setuptools import setup

setup(name='okmesopy',
      version='1.1.1',
      description='Tools to download and manipulate data from OK Mesonet',
      url='https://github.com/bjrubinstein/okmesopy',
      author='Ben Rubinstein',
      author_email='brubinst@hawk.iit.edu',
      license='GNU',
      packages=['okmesopy'],
      install_requires=['numpy',
                        'pandas',
                        'pyshp',
                        'pyproj',
                        'geopy',
                        'missingno',
                        ],
      keywords=['Mesonet','climatology','meteorology','hydrology'],
      zip_safe=False)
